#!/usr/bin/perl

########################################################################
# metaf.pl 1.40
#   (CGI) script to fetch, decode, and display METAR/TAF/SYNOP messages
#
# copyright (c) metaf2xml 2006-2010
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

=head1 NAME

metaf.pl

=head1 SYNOPSIS

as a standalone script used from the command line:

  metaf.pl 'type_metaf=...&type_synop=...&lang=...&format=...&src_metaf=...&src_synop=...&mode=...&hours=...&msg_metaf=...&msg_synop=...'

as a CGI script used by an HTML form (HTTP method POST):

  <form method="POST" action="metaf.pl">
   <input name="type_metaf" ... >
   <input name="type_synop" ... >
   <input name="lang" ... >
   <input name="format" ... >
   <input name="src_metaf" ... >
   <input name="src_synop" ... >
   <input name="mode" ... >
   <input name="hours" ... >
   <input name="msg_metaf" ... >
   <input name="msg_synop" ... >
  </form>

=head1 DESCRIPTION

This script can fetch up-to-date or historical weather information (METAR, TAF,
or SYNOP messages) from an Internet server or a local file, decode weather
messages, and write them as XML or HTML.
It can be used with GET or POST requests if it is installed to be called by a
web server, or from the command line.

=head1 DEPENDENCIES

The Perl module C<CGI>, the function C<POSIX::mktime> from the Perl module
C<POSIX>, and the program C<metaf2xml.pl> are required.

The program C<curl> is required if the script is to fetch data from an Internet
server.
The program C<metafsrc2raw.pl> is required if data is to be fetched from an
Internet server with file based access.

The program C<xsltproc> is required if the script is to convert the XML to HMTL
or text.

=head1 ABBREVIATIONS

=over

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item adds

the Aviation Digital Data Service of the NWS

=item addsds

the experimental data server of the Aviation Digital Data Service of the NWS

=item cod

College of DuPage

=item fsu

Florida State University

=back

=head1 PARAMETERS

The following CGI parameters are recognised:

=over

=item type_metaf

type of content in C<msg_metaf>: B<metar> (default), B<taf>, B<icao> (ICAO id(s)
to fetch)

=item type_synop

type of content in C<msg_synop>: B<synop> (default), B<wmo> (WMO id(s) to fetch)

=item lang

language for output of script and display of parsed message: B<de> (default),
B<en>, B<es>, B<ru>

=item format

output format: B<html> (default), B<text>, or B<xml> (with style sheet
reference)

=item src_metaf

source for METAR/TAF messages to be fetched:
B<noaa> (NOAA/IWS)[1] (default), B<adds>[2], B<addsds>[2], B<ogimet>[2],
B<nws>[3], B<cod>[3] (no TAF), B<fsu>[3] (no TAF),
B<file> (forced automatically if F<$CGI_DATA_DIR/metaf.txt> exists)

Note: B<nws>, B<cod>, and B<fsu> are not available in the web based interfaces.

=item src_synop

source for SYNOP messages to be fetched:
B<nws_land>[3] (default), B<nws_ship>[3], B<ogimet>[2], B<cod>[3], B<fsu>[3],
B<file> (forced automatically if F<$CGI_DATA_DIR/synop.txt> exists)

Note: B<fsu> is not available in the web based interfaces.

=item mode

which messages to fetch and how to display: B<latest>/detailed (default),
all/B<summary>

=item hours

for mode B<summary>: number of hours to fetch messages for (0..72, default: 24)

Note: The exact time range for which data is received depends on the source of
messages.

=item msg_metaf

METAR/TAF message to be processed, or ICAO id(s)

=item msg_synop

SYNOP message to be processed, or WMO id(s)

=back

[1]: if mode is B<latest>, see [2]; if mode is B<summary>, see [3]

[2]: the HTML form provided by the server will be used

[3]: files will be fetched from the server and searched

=cut

# all times are UTC (esp. required for mktime())
$ENV{TZ} = 'GMT0';

use POSIX qw(mktime);
use CGI qw(:standard -no_xhtml);

#
# restrict use of CGI:
#

# uploads are not allowed
$CGI::DISABLE_UPLOADS = 1;

=pod

The length of the body of the POST request (not the length of query string of
the URL) is restricted to 600 bytes.

=cut

$CGI::POST_MAX = 600;

=pod

The following directories and files are used by the script and can be adapted to
local needs by modifying the script (e.g. during install):

=over

=item F<$METAF2XML_BIN>

directory where C<metaf2xml.pl> and C<metafsrc2raw.pl> are expected (default:
F</opt/metaf2xml/bin>)

=item F<$METAF2XML_XSL>

directory where the XSL scripts are located (default: F</opt/metaf2xml/share>)

=item F<$CGI_DATA_DIR>

directory for the local copies of messages (default: F</tmp>)

=item F<$CURL>

path and options for C<curl> (default: C<curl -sSf --stderr ->)

=item F<$CURL_PROXY_OPTS>

options for C<curl> if a HTTP proxy should be used (default: empty)

=back

=cut

########################################################################
# config section, possibly changed by install
########################################################################
my $METAF2XML_BIN = '/opt/metaf2xml/bin';
my $METAF2XML_XSL = '/opt/metaf2xml/share';
my $CGI_DATA_DIR = '/tmp';
my $CGI_LOG_FILE = '/tmp/metaf2xml.log';
my $CURL = 'curl -s0Sf --stderr -';
my $CURL_PROXY_OPTS = '';

my $re_report_type_synop = '(?:S[IMN])';
my $re_report_type_metaf = '(?:S[AP]|F[CT])';
my $re_header_synop = '(?:AAXX .{5}|(?:BB|OO)XX)';
my $re_ICAO = '(?:[A-Z][A-Z\d]{3})';

#
# make a list of strings unique
#
sub mk_uniq {
    my $uniq = '';
    for (@_) {
        if ($uniq eq '') {
            $uniq = $_;
        } elsif (index(" $uniq ", " $_ ") == -1) {
            $uniq .= " $_";
        }
    }
    return $uniq;
}

#
# get the contents of the directory on the NOAA/NWS server from the server
#   type: nws_metar, nws_taf, nws_land, nws_ship
sub files_nws {
    my ($type, $hours) = @_;
    my ($url, $response, @files);

    # fetch directory listing
    # also:      ftp://tgftp.nws.noaa.gov/SL.us008001/DF.an/
    $url =   'http://weather.noaa.gov/pub/SL.us008001/DF.an/'
           . {
               nws_metar => 'DC.sflnd/DS.metar',
               nws_taf   => 'DC.sflnd/DS.tafst',
               nws_land  => 'DC.sflnd/DS.synop',
               nws_ship  => 'DC.sfmar/DS.ships',
             }->{$type};

    open(FP, "$CURL $CURL_PROXY_OPTS $url/ls-lt 2>&1 || echo ERROR |")
        || return "ERROR - " . uc "$url/ls-lt - $!";

    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return "ERROR - " . uc "$url/ls-lt - $response";
    }

    @files = $response =~ / (sn.\d{4}.txt)/g;

    # don't use the oldest file, it may have/may shortly become the newest
    $#files--;

    # truncate array if less hours are requested
    # (one file every 5 min. = 12 per hour)
    $#files = $hours * 12 if $hours * 12 < $#files;

    return $url, @files;
}

#
# determine file names for data from NOAA/IWS (cycle) files
#   type: metar_cycle_iws, staf_cycle_iws, taf_cycle_iws
sub files_cycled_iws {
    my ($type, $hours) = @_;
    my ($hour, @files);

    if ($type eq 'metar_cycle_iws') {
        # METAR cycle files start 15 min. before the hour
        $hour = (gmtime(time() + 15 * 60))[2];
    } else {
        # (S)TAF cycle files start 1 hour before the main synoptic hour
        $hour = (gmtime(time() + 60 * 60))[2];
    }
    for (0 .. ($hours > 23 ? 23 : $hours)) {
        push @files, sprintf('%02dZ.TXT', $hour)
            if $type eq 'metar_cycle_iws' || $hour % 6 == 0;
        $hour--;
        $hour = 23 if $hour == -1;
    }
    return   'http://weather.noaa.gov/pub/data/'
           . { metar_cycle_iws => 'observations/metar',
               staf_cycle_iws  => 'forecasts/shorttaf',
               taf_cycle_iws   => 'forecasts/taf',
             }->{$type}
           . '/cycles',
           @files;
}

#
# determine file names for data from College of DuPage
#   type: syn, sao
sub files_cod {
    my ($type, $hours) = @_;
    my ($now, @files);

    $now = time();

    # one file every hour
    for (0..$hours) {
        my @date_arr;

        @date_arr = (gmtime($now - $_ * 60 * 60))[5, 4, 3, 2];
        $date_arr[0] %= 100; # year is offset to 1900, need 2 digits, only
        $date_arr[1]++;      # month starts from 0
        push @files, sprintf("%02d%02d%02d%02d.$type", @date_arr);
    }
    return 'http://weather.cod.edu/digatmos/' . $type, @files;
}

#
# determine file names for data from Florida State University
#   type: syn, metar
sub files_fsu {
    my ($type, $hours) = @_;
    my ($now, @files);

    $now = time();

    # one file every hour
    for (0..$hours) {
        my (@date_arr, $date_str);

        @date_arr = (gmtime($now - $_ * 60 * 60))[5, 4, 3, 2];
        $date_arr[0] += 1900; # year is offset to 1900, need 4 digits
        $date_arr[1]++;       # month starts from 0
        $date_str = sprintf "%04d%02d%02d%02d", @date_arr;
        push @files, substr($date_str, 0, 8) . "/$date_str.$type";
    }
    return 'http://dweb.met.fsu.edu/rawdata/' . $type, @files;
}

#
# scan files with SYNOP data for $wmoid
#
sub search_files_synop {
    my ($wmoid, $mode, $src, $hours) = @_;
    my ($url, @files, @wmoids, $no_wmoid, @arr);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    if ($src eq 'file') {
        ($url, @files) = ('', "$CGI_DATA_DIR/synop.txt");
    } elsif ($src eq 'cod') {
        ($url, @files) = files_cod('syn', $hours);
    } elsif ($src eq 'fsu') {
        ($url, @files) = files_fsu('syn', $hours);
    } elsif ($src eq 'nws_land' || $src eq 'nws_ship') {
        ($url, @files) = files_nws($src, $hours);
        return $url if $url =~ /ERROR/;
        $src = 'nws';
    } else {
        return;
    }

    $wmoid =~ s/ /|/g;
    @wmoids = split /[|]/, $wmoid;
    $no_wmoid = $#wmoids + 1;
    @arr = ();
    for my $file (@files) {
        my ($filename, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $filename = "($CURL $CURL_PROXY_OPTS $url/$file 2>&1 || echo ERROR) | $METAF2XML_BIN/metafsrc2raw.pl -H1 -NFsynop_$src |";
        } else {
            $filename = $file;
        }

        open(FP, $filename)
            || return "$wmoid - ERROR - " . uc(($url || $file) . " - $!");
        @file_arr = ();
        $msg_seq  = '';
        while (<FP>) {
            my ($line, $report_type, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            $line = $_;

            ($report_type, $msg) = $line =~ /^(..) (.*)/;
            next unless defined $report_type && defined $msg;

            $id = ($msg =~ m@^(?:AAXX .{5}|(?:BB|OO)XX) ([A-Z\d/]{3,})@)[0];
            next unless defined $id;

            $report_type_id = "$report_type-$id";
            $report_type_matches = index("|$wmoid|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$wmoid|", "|$id|") > -1;

            # looking for messages with this id?
            if ($report_type_matches) {
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                $report_type_allowed = $re_report_type_synop;
            } else {
                next;
            }

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @file_arr, $_
                    unless grep { /^$line$/ } (@arr, @file_arr);
                next;
            }

            # found message in previous file(s)?
            next if grep { /^$report_type_allowed $re_header_synop $id / } @arr;

            # request  : input             => output
            # id       : Sy-id,Sx-id,Sz-id => Sz-id
            # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
            # Sx-id,id : Sx-id,Sy-id       => Sx-id,Sy-id
            # Sx-id,id : Sy-id,Sx-id       => Sx-id
            # looking for messages with the report type of this message?
            if ($report_type_matches) {
                if ($msg_found{$report_type_id}) {  # record type found already?
                    $msg_seq =~ s/ $report_type_id //;   # remove existing entry
                } else {
                    $no_wmoid--;                    # mark record type as found
                }
                $msg_seq = " $report_type_id " . $msg_seq;   # prepend new entry
                $msg_found{$report_type_id} = $line;         # save new value
            }

            # looking for messages with any report type?
            if ($report_type_any) {
                if ($msg_found{"Sx-$id"}) {
                    $msg_seq =~ s/ Sx-$id //;
                } else {
                    $no_wmoid--;
                }
                # prepend only if not done already above
                $msg_seq = " Sx-$id " . $msg_seq
                    unless $report_type_matches;
                $msg_found{"Sx-$id"} = $line;
            }
        }
        close FP;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_wmoid == 0;
    }
    return @arr;
}

#
# scan files with METAR/TAF data for $icao
#
sub search_files_metaf {
    my ($icao, $mode, $src, $hours) = @_;
    my ($url, @files, @icaos, $no_icao, @arr);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    if ($src eq 'file') {
        ($url, @files) = ('', "$CGI_DATA_DIR/metaf.txt");
    } elsif ($src eq 'cod') {
        ($url, @files) = files_cod('sao', $hours);
        $src = 'metaf_cod';
    } elsif ($src eq 'fsu') {
        ($url, @files) = files_fsu('metar', $hours);
        $src = 'metaf_fsu';
    } elsif ($src eq 'nws_metar' || ($src eq 'nws_taf' && $mode ne 'summary')) {
        ($url, @files) = files_nws($src, $hours);
        return $url if $url =~ /ERROR/;
        $src = 'metaf_nws';
    } elsif (   $src eq 'metar_cycle_iws'
             || ($src =~ /^s?taf_cycle_iws$/ && $mode ne 'summary'))
    {
        ($url, @files) = files_cycled_iws($src, $hours);
        return $url if $url =~ /ERROR/;
    } else {
        return '';
    }

    $icao =~ s/ /|/g;
    @icaos = split /[|]/, $icao;
    $no_icao = $#icaos + 1;
    @arr = ();
    for my $file (@files) {
        my ($filename, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $filename = "($CURL $CURL_PROXY_OPTS $url/$file 2>&1 || echo ERROR) | $METAF2XML_BIN/metafsrc2raw.pl -H1 -NF$src |";
        } else {
            $filename = $file;
        }

        open(FP, $filename)
            || return "$icao - ERROR - " . uc(($url || $file) . " - $!");
        @file_arr = ();
        $msg_seq  = '';
        while (<FP>) {
            my ($line, $report_type, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            return $_ if $_ =~ /ERROR/;
            $line = $_;

            ($report_type, $msg) = /^(..) (.*)/;
            next unless defined $report_type && defined $msg;

            $id = ($msg =~ m@^(?:COR |AMD )*($re_ICAO) @)[0];
            next unless defined $id;

            $report_type_id = "$report_type-$id";
            $report_type_matches = index("|$icao|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$icao|", "|$id|") > -1;

            # looking for messages with this id?
            if ($report_type_matches) {
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                $report_type_allowed = $re_report_type_metaf;
            } else {
                next;
            }

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @file_arr, $_
                    unless grep { /^$line$/ } (@arr, @file_arr);
                next;
            }

            # found message in previous file(s)?
            next if grep { /^$report_type_allowed (?:COR |AMD )*$id / } @arr;

            # looking for messages with the report type of this message?
            if ($report_type_matches) {
                if ($msg_found{$report_type_id}) {  # record type found already?
                    $msg_seq =~ s/ $report_type_id //;   # remove existing entry
                    $msg_seq =~ s/ SP-$id //             # SA: also remove SP
                        if $report_type eq 'SA';
                } else {
                    $no_icao--;                   # mark record type as found
                    $no_icao--                    # SA: don't search for more SP
                        if $report_type eq 'SA' && !$msg_found{"SP-$id"};
                }
                $msg_seq = " $report_type_id " . $msg_seq;   # prepend new entry
                $msg_found{$report_type_id} = $line;         # save new value
                $msg_found{"SP-$id"} = $line                 # SA: mark SP found
                    if $report_type eq 'SA';
            }
        }
        close FP;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_icao == 0;
    }
    return @arr;
}

#
# query OGIMET
#
sub query_ogimet {
    my ($id, $mode, $which, $hours, $type) = @_;
    my (@date_inicio, @date_fin, $url, $response);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    @date_inicio = (gmtime(time() - $hours * 3600))[5, 4, 3, 2];
    $date_inicio[0] += 1900;              # year is offset to 1900
    $date_inicio[1]++;                    # month starts from 0
    @date_fin = (gmtime())[5, 4, 3, 2];
    $date_fin[0] += 1900;                 # year is offset to 1900
    $date_fin[1]++;                       # month starts from 0
    $url = sprintf 'http://www.ogimet.com/display_%ss.php?tipo=%s&ord=REV&nil=SI&fmt=txt&ano=%d&mes=%02d&day=%02d&hora=%02d&anof=%d&mesf=%02d&dayf=%02d&horaf=%02d&lugar=',
           $which, $type, @date_inicio, @date_fin;

    open(FP, "$CURL $CURL_PROXY_OPTS '$url$id' 2>&1 || echo ERROR |")
        || return "$id - ERROR - " . uc "$url - $!";

    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return "$id - ERROR - " . uc "$url - $response";
    }
    return $response;
}

#
# query OGIMET for $wmoid
#
sub query_synop_ogimet {
    my ($wmoid, $mode, $wmoid_uniq, $hours, $type) = @_;
    my ($response, @arr, %msg_found);
    my $re_day = '(?:[0-2]\d|3[01])';
    my $re_hour = '(?:[01]\d|2[0-3])';

    $wmoid_uniq =~ s/ /+/g;
    $response = query_ogimet $wmoid_uniq, $mode, 'synop', $hours, $type;
    return $response if $response =~ /ERROR/;

    $wmoid =~ s/ /|/g;

    $_ = (split /<[\/]?pre>/, $response)[1];     # extract answer
    s/^#.*//mg;                                  # make comment lines empty
    s/\n+/\n/g;                                  # delete consecutive newlines
    s/\n +/ /g;                                  # delete newlines in messages
    s/=//g;                                      # delete EOM
    s/^.{13}//mg;                                # remove time stamps
    for (split /\n/) {                           # process each message
        my ($id, $report_time, $report_type, $was_added);
        my ($report_type_id, $report_type_matches, $report_type_any);

        if (m@^AAXX $re_day($re_hour)[0134/] ([A-Z\d]{3,}) @) {
            $report_time = $1;
            $id = $2;
        } elsif (m@^(?:BB|OO)XX ([A-Z\d]{3,}) $re_day($re_hour)[0134/] @) {
            $id = $1;
            $report_time = $2;
        }
        next unless $2;

        if ($report_time =~ /00|06|12|18/) {
            $report_type = 'SM';
        } elsif ($report_time =~ /03|09|15|21/) {
            $report_type = 'SI';
        } else {
            $report_type = 'SN';
        }

        $report_type_id = "$report_type-$id";
        $report_type_matches = index("|$wmoid|", "|$report_type_id|") > -1;
        $report_type_any     = index("|$wmoid|", "|$id|") > -1;

        if ($mode eq 'summary') {
            push @arr, "$report_type $_"
                if $report_type_matches || $report_type_any;
            next;
        }

        # request  : input             => output
        # id       : Sy-id,Sx-id,Sz-id => Sy-id
        # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
        # Sx-id,id : Sx-id,Sy-id       => Sx-id
        # Sx-id,id : Sy-id,Sx-id       => Sy-id,Sx-id
        $was_added = 0;
        if ($report_type_matches && !exists $msg_found{$report_type_id}) {
            $msg_found{$report_type_id} = undef;
            $was_added = 1;
            push @arr, "$report_type $_";
        }
        if ($report_type_any && !exists $msg_found{"Sx-$id"}) {
            $msg_found{"Sx-$id"} = undef;
            push @arr, "$report_type $_"
                unless $was_added;
        }
    }
    return @arr;
}

#
# query NOAA/IWS for $icao
#
sub query_metaf_iws {
    my ($icao, $type, $mode) = @_;
    my ($response, $url, $msg_type);

    return () if $mode eq 'summary' && $type ne 'metar';
    return () if $icao eq '';

    $icao = mk_uniq $icao =~ /(?:$re_report_type_metaf-)([^ ]+)/og;
    $icao =~ s/ /+/g;

    $url = 'http://weather.noaa.gov/cgi-bin/mget' . $type . '.pl?cccc=';

    open(FP, "$CURL $CURL_PROXY_OPTS $url'$icao' 2>&1 || echo ERROR |")
        || return "$icao - ERROR - " . uc "$url - $!";

    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $msg_type = { metar => 'SA', staf => 'FC', taf => 'FT' }->{$type};
    $response =~ s/[\n\r]+/ /g;
    $response =~ s/ +/ /g;
    $response = join("\n", $response =~ /<font face=.couri[^>]*?> *(?:<pre>)? *((?:METAR|SPECI|TAF)? *[^<]+)/g);
    $response =~ s/ $//mg;
    $response =~ s/^(?:TAF |METAR )*/$msg_type /mg;
    $response =~ s/^SA SPECI /SP /mg;
    return split /\n/, $response;
}

#
# query NOAA/ADDS for $icao
#
sub query_metaf_adds {
    my ($icao, $mode, $hours, $need_taf) = @_;
    my ($options, $response, $url, $data);

    # only integer hours allowed
    $options = $mode eq 'summary' ? 'past ' . int($hours) . ' hours'
                                  : (  'most recent only'
                                     . ($need_taf ? '&chk_tafs=on' : ''));
    $icao =~ s/ /+/g;

    $url  = 'http://adds.aviationweather.gov/metars/';
    $data = "std_trans=standard&chk_metars=on&hoursStr=$options&submitmet=Submit&station_ids=";

    open(FP, "$CURL $CURL_PROXY_OPTS -d '$data$icao' $url 2>&1 || echo ERROR |")
        || return "$icao - ERROR - " . uc "$url - $!";

    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $response =~ s/[\n\r]+//g;
    $response =~ s/ +/ /g;
    $response =~ s/(<font face=.mono[^>]*?>)/$1SA /ig;
    $response =~ s/(<font face=.mono[^>]*?>SA SPECI )/$1SP /ig;
    $response =~ s/(<pre>)/$1FT /ig;
    return $response =~ /(?:<font face=.mono[^>]*?>|<pre>)([^<]+)/ig;
}

#
# query NOAA/ADDS dataserver for $icao
#
sub query_metaf_addsds {
    my ($icao, $type, $mode, $hours) = @_;
    my ($response, $url);
    my (%issue_time, @raw_text, %metar, %taf);

    return () if $mode eq 'summary' && $type ne 'metar';
    return () if $icao eq '';

    $icao =~ s/ /,/g;
    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest

    $url = 'http://weather.aero/dataserver_current/httpparam?requestType=retrieve&format=xml&dataSource=';
    if ($type eq 'metar') {
        $url .= 'metars&fields=raw_text,station_id,observation_time,metar_type&hoursBeforeNow=' . $hours;
    } else {
        $url .= 'tafs&fields=raw_text,station_id,issue_time,valid_time_from,valid_time_to&hoursBeforeNow=0&timeType=valid';
    }
    $url .= '&stationString=';

    open(FP, "$CURL $CURL_PROXY_OPTS '$url$icao' 2>&1 || echo ERROR |")
        || return "$icao - ERROR - " . uc "$url - $!";

    {
        local $/;
        $response = <FP>;
    }
    close FP;
    if ($response =~ /ERROR/) {
        $response =~ s/[\n\r]+/ /g;
        return ("$icao - ERROR - " . uc "$url - $response");
    }
    $response =~ s/[\n\r]+/ /g;
    $response =~ s/ +/ /g;
    if ($type eq 'metar') {
        # copy to hash, key is observation_time
        for ($response =~ /<((?:raw_text|station_id|observation_time|metar_type)>[^<]+)/ig)
        {
            /^([^>]+).(.*)/;
            $metar{$1} = $2;
            if ($1 eq 'metar_type') {
                if (scalar keys %metar eq 4) {
                    my $key = $mode eq 'summary'
                              ? $metar{raw_text}
                              : "$metar{station_id},$metar{metar_type}";
                    if (   !exists $issue_time{$key}
                        || $issue_time{$key} lt $metar{observation_time})
                    {
                        $issue_time{$key} = $metar{observation_time};
                        $metar{raw_text} =~ s/^(?:METAR |SPECI )*//;
                        $metar{raw_text} =
                            { METAR=>'SA ', SPECI=>'SP ' }->{$metar{metar_type}}
                          . $metar{raw_text};

                        if ($metar{metar_type} eq 'METAR' || $mode eq 'summary')
                        {
                            push @raw_text, $metar{raw_text};
                        # not summary: add SPECI only if there is no METAR yet
                        } elsif (!exists $issue_time{"$metar{station_id},METAR"})
                        {
                            push @raw_text, $metar{raw_text};
                        }
                    }
                }
                %metar = ();
            }
        }
        for ($response =~ /(<error>[^<]+)/ig) {
            /^([^>]+).(.*)/;
            push @raw_text, "$icao - ERROR - " . uc $2;
        }
        return @raw_text;
    }

    # copy to hash, key is issue_time
    for ($response =~ /<((?:raw_text|station_id|issue_time|valid_time_from|valid_time_to)>[^<]+)/ig)
    {
        /^([^>]+).(.*)/;
        $taf{$1} = $2;
        if ($1 eq 'valid_time_to') {
            if (scalar keys %taf eq 5) {
                my ($hours, $msg_type, @date_arr);

                @date_arr[5,4,3,2,1,0] = split /[-TZ:]/, $taf{valid_time_from};
                $date_arr[5] -= 1900;                   # year is offset to 1900
                $date_arr[4]--;                         # month starts from 0
                $taf{valid_time_from} = mktime @date_arr, -1, -1, -1;
                @date_arr[5,4,3,2,1,0] = split /[-TZ:]/, $taf{valid_time_to};
                $date_arr[5] -= 1900;                   # year is offset to 1900
                $date_arr[4]--;                         # month starts from 0
                $taf{valid_time_to} = mktime @date_arr, -1, -1, -1;
                $hours = ($taf{valid_time_to} - $taf{valid_time_from}) / 3600;
                $msg_type = $hours >= 12 ? 'FT' : 'FC';
                my $key = "$taf{station_id},$msg_type";
                if (   !exists $issue_time{$key}
                    || $issue_time{$key} lt $taf{issue_time})
                {
                    $issue_time{$key} = $taf{issue_time};
                    push @raw_text, "$msg_type $taf{raw_text}";
                }
            }
            %taf = ();
        }
    }

    return @raw_text;
}

#
# query OGIMET for $icao
#
sub query_metaf_ogimet {
    my ($icao, $mode, $hours, $type) = @_;
    my ($response, @arr, %msg_found, $taf_type);

    $icao =~ s/ /+/g;
    # also returns TAF!
    $response = query_ogimet $icao, $mode, 'metar', $hours, $type;
    return $response if $response =~ /ERROR/;

    $_ = (split /<[\/]?pre>/, $response)[1];  # extract answer
    s/^#  TAF/1234567890123TAF/mg;            # save TAF type
    s/^#.*//mg;                               # make comment lines empty
    s/\n+/\n/g;                               # delete consecutive newlines
    s/\n +/ /g;                               # delete newlines in messages
    s/=//g;                                   # delete EOM
    s/^.{13}//mg;                             # remove time stamps
    $taf_type = '';
    for (split /\n/) {                        # process each message
        if (/^TAF CORTOS/) {
            $taf_type = 'FC';
            next;
        }
        if (/^TAF LARGOS/) {
            $taf_type = 'FT';
            next;
        }

        s/^SPECI /SP /;
        s/^METAR /SA /;
        s/^TAF/$taf_type/;
        /^($re_report_type_metaf) (?:COR |AMD )*($re_ICAO) /;
        next unless defined $1;
        if ($mode eq 'summary' || !exists $msg_found{"$1-$2"}) {
            push @arr, $_;
            $msg_found{"$1-$2"} = undef;
        }
    }
    return @arr;
}

=pod

If there are any CGI errors the type of error is reported and the script
terminates.

=cut

open STDERR, ">&STDOUT";

charset('UTF-8');

if (cgi_error()) {
    print header({ status => cgi_error() }) .
        start_html({ title => cgi_error() }) . "\n" .
        b(cgi_error()) . "\n" .
        end_html();
    exit 0;
}

=pod

All parameters are checked to have one of the allowed values.

=cut

#
# determine type_metaf: check request parameter. default: metar
#
my $type_metaf = param('type_metaf');
$type_metaf = 'metar'
    unless defined $type_metaf && ($type_metaf eq 'taf' || $type_metaf eq 'icao');

#
# determine type_synop: check request parameter. default: synop
#
my $type_synop = param('type_synop');
$type_synop = 'synop'
    unless defined $type_synop && $type_synop eq 'wmo';

=pod

If the parameter C<lang> is not set the first allowed value of the HTTP header
C<Accept-Language> is used.

=cut

#
# determine language: check request parameter, then browser header. default: de
#
my $lang = param('lang');
if (!defined $lang) {
    my $accept_lang = http('Accept-Language');
    if (defined $accept_lang) {
        for (split(/,/, $accept_lang)) {
            last if $_ =~ /^de/i;
            if ($_ =~ /^en/i) {
                $lang = 'en';
                last;
            }
            if ($_ =~ /^es/i) {
                $lang = 'es';
                last;
            }
            if ($_ =~ /^ru/i) {
                $lang = 'ru';
                last;
            }
        }
    }
}
$lang = 'de'
    unless defined $lang && ($lang eq 'en' || $lang eq 'es' || $lang eq 'ru');

#
# determine format: check request parameter, then browser header. default: html
#
my $format = param('format');
$format = 'html'
    unless defined $format && ($format eq 'text' || $format eq 'xml');
if ($format eq 'xml') {
    my $accept = http('Accept');
    if ($accept) {
        my $accept_all = ",$accept," =~ m@[ ,]\*/\*[;,]@;
        my $accept_xml = ",$accept," =~ m@[ ,](?:text|application)/(?:[a-z+-]*xml[a-z+-]*|\*)[;,]@;
        my $accept_xsl = ",$accept," =~ m@[ ,](?:text|application)/(?:[a-z+-]*x[sm]l[a-z+-]*|\*)[;,]@;
        $format = 'html' unless $accept_all || ($accept_xml && $accept_xsl);
    }
}

#
# determine src_synop, src_metaf: check request parameter
#
my ($src_metaf, $src_synop);
if (-f "$CGI_DATA_DIR/metaf.txt") {
    $src_metaf = 'file';
} else {
    $src_metaf = param('src_metaf');
    $src_metaf = 'noaa'
        unless    defined $src_metaf
               && $src_metaf =~ /^(?:adds|addsds|ogimet|nws|cod|fsu)$/;
}
if (-f "$CGI_DATA_DIR/synop.txt") {
    $src_synop = 'file';
} else {
    $src_synop = param('src_synop');
    $src_synop = 'nws_land'
        unless    defined $src_synop
               && $src_synop =~ /^(?:nws_ship|ogimet|cod|fsu)$/;
}

#
# determine mode: check request parameter. default: latest
#
my $mode = param('mode');
$mode = 'latest' unless defined $mode && $mode eq 'summary';

#
# determine number of hours (for mode summary): check request param. default: 24
#
my $hours = param('hours');
$hours = 24 unless defined $hours && $hours >= 0 && $hours <= 72;

=pod

The parameters C<msg_metaf> and C<msg_synop> are truncated to 500 characters.
Letters are converted to upper case. Any characters that are not allowed in a
message are removed.

=cut

my $msg_metaf = param('msg_metaf') ? substr(param('msg_metaf'), 0, 500) : '';
$msg_metaf = uc $msg_metaf;
$msg_metaf =~ s/[^ -~]/?/gos; # avoid invalid XML and HTML

my $msg_synop = param('msg_synop') ? substr(param('msg_synop'), 0, 500) : '';
$msg_synop = uc $msg_synop;
$msg_synop =~ s/[^ -~]/?/gos; # avoid invalid XML and HTML

my $is_cmdline = server_software() eq 'cmdline';
my $need_html  = $is_cmdline && $format eq 'text' ? 0 : 1;

#
# log CGI request unless called from the command line
#
if (!$is_cmdline && open(LOG, '>>', $CGI_LOG_FILE)) {
    # garble IP address, but still allow to group the requests
    my $addr = $ENV{'HTTP_X_REMOTE_ADDR'} || remote_addr();
    $addr =~ s/\.\d+/.x/;
    $addr =~ s/\d+$/x/;
    print LOG gmtime() . ' ' . $addr . ' ' . request_method()
              . " $type_metaf $type_synop $lang $format $src_metaf $src_synop $mode $hours" . (defined param('gui') ? ' gui' : '') . " '$msg_metaf' '$msg_synop'\n";
    close LOG;
}

=pod

If the METAR/TAF message starts with C<METAR>, C<SPECI>, or C<TAF>, the
parameter C<type_metaf> is ignored.

=cut

#
# overwrite type_metaf if specified in message
#
$type_metaf = 'metar' if $msg_metaf =~ /^(?:METAR|SPECI) /;
$type_metaf = 'taf' if $msg_metaf =~ /^TAF /;

=pod

If the SYNOP message starts with C<SYNOP>, the parameter C<type_synop> is
ignored.

=cut

#
# overwrite type_synop if specified in message
#
$type_synop = 'synop' if $msg_synop =~ /^SYNOP /;

#
# translated strings
#
my %trans = (
    title_en => 'metaf2xml: METAR/TAF/SYNOP decoder',
    title_de => 'metaf2xml: METAR/TAF/SYNOP-Übersetzer',
    title_es => 'metaf2xml: METAR/TAF/SYNOP traductor',
    title_ru => 'metaf2xml: МЕТАР/ТАФ/СИНОП декодер',
    header_en => 'metaf2xml: METAR/TAF/SYNOP decoder',
    header_de => 'metaf2xml: METAR/TAF/SYNOP-Übersetzer',
    header_es => 'metaf2xml: METAR/TAF/SYNOP traductor',
    header_ru => 'metaf2xml: МЕТАР/ТАФ/СИНОП декодер',
    infosrc_en => 'A good source for up-to-date METAR and TAF messages is',
    infosrc_de => 'Eine gute Quelle für aktuelle METAR- und TAF-Meldungen ist',
    infosrc_es => 'Una buena fuente por mensajes actuales de METAR y TAF es',
    infosrc_ru => 'Хороший источник для современных сообщений МЕТАР и ТАФ',
    infofetch_en => 'If the option "ICAO id(s)" or "WMO id(s)" is chosen and one or more Station-Ids are given, messages will be fetched from the selected source (or if this option is disabled from a file on the server where this web page comes from).',
    infofetch_de => 'Wird die Option "ICAO Id(s)" oder "WMO Id(s)" gewählt und ein oder mehrere Stations-Kürzel angegeben, werden Meldungen von der ausgewählten Quelle (oder, falls diese Option gesperrt ist, von einer Datei auf dem Server, von dem diese Web-Seite kommt) geholt.',
    infofetch_es => 'Si es elegido la opción "código(s) OACI" o "código(s) OMM" y son dado uno o mas códigos de la estación, los mensajes serán traído de la fuente elegida (o si esta opción es minusválida de un fichero del servidor de donde viene este página web).',
    infofetch_ru => 'Если выбор "ICAO ид" или "ВМО ид" выбран, и одни или более ид станции даются, сообщения принесены из отобранного источника (или если этот выбор инвалиды от файла на сервере, откуда эта веб-страница прибывает).',
    example_en => 'example messages',
    example_de => 'Beispiele für Meldungen',
    example_es => 'ejemplos',
    example_ru => 'примеры',
    type1_en => 'METAR message or', type2_en => 'TAF message or',
     type3_en => 'SYNOP message or',
     type4_en => 'ICAO id(s)', type5_en => 'WMO id(s)',
    type1_de => 'METAR Meldung oder', type2_de => 'TAF Meldung oder',
     type3_de => 'SYNOP Meldung oder',
     type4_de => 'ICAO Id(s)', type5_de => 'WMO Id(s)',
    type1_es => 'METAR mensaje o', type2_es => 'TAF mensaje o',
     type3_es => 'SYNOP mensaje o',
     type4_es => 'código(s) OACI', type5_es => 'código(s) OMM',
    type1_ru => 'МЕТАР сообщение или', type2_ru => 'ТАФ сообщение или',
     type3_ru => 'СИНОП сообщение или',
     type4_ru => 'ICAO ид', type5_ru => 'ВМО ид',
    source_en   => 'fetch messages from',
    source1_en  => 'NOAA/IWS', source2_en => 'NOAA/ADDS',
     source3_en => 'NOAA/ADDS dataserver', source4_en => 'NOAA/NWS land',
     source5_en => 'College of DuPage', source6_en => 'OGIMET',
     source7_en => 'NOAA/NWS ship',
    source_de   => 'hole Meldungen von',
    source1_de  => 'NOAA/IWS', source2_de => 'NOAA/ADDS',
     source3_de => 'NOAA/ADDS dataserver', source4_de => 'NOAA/NWS land',
     source5_de => 'College of DuPage', source6_de => 'OGIMET',
     source7_de => 'NOAA/NWS ship',
    source_es   => 'trae mensajes de',
    source1_es  => 'NOAA/IWS', source2_es => 'NOAA/ADDS',
     source3_es => 'NOAA/ADDS dataserver', source4_es => 'NOAA/NWS land',
     source5_es => 'College of DuPage', source6_es => 'OGIMET',
     source7_es => 'NOAA/NWS ship',
    source_ru   => 'усилия сообщения от',
    source1_ru  => 'NOAA/IWS', source2_ru => 'NOAA/ADDS',
     source3_ru => 'NOAA/ADDS dataserver', source4_ru => 'NOAA/NWS land',
     source5_ru => 'College of DuPage', source6_ru => 'OGIMET',
     source7_ru => 'NOAA/NWS ship',
    language_en => 'language',
    language_de => 'Sprache',
    language_es => 'lengua',
    language_ru => 'язык',
    format_en   => 'output as',
    format1_en  => 'HTML', format2_en => 'text',
     format3_en => 'XML (+XSL -> HTML)',
    format_de   => 'gib aus als',
    format1_de  => 'HTML', format2_de => 'Text',
     format3_de => 'XML (+XSL -> HTML)',
    format_es   => 'presenta como',
    format1_es  => 'HTML', format2_es => 'texto',
     format3_es => 'XML (+XSL -> HTML)',
    format_ru   => 'показа в',
    format1_ru  => 'HTML', format2_ru => 'текст',
     format3_ru => 'XML (+XSL -> HTML)',
    mode_en     => 'fetch/display',
    mode1_en    => 'latest/detailed', mode2_en => 'all/summary since',
    mode_de     => 'hole/zeige',
    mode1_de    => 'letzte/detailliert', mode2_de => 'alle/Übersicht seit',
    mode_es     => 'trae/presenta',
    mode1_es    => 'último/detallado', mode2_es => 'todos/resumen desde',
    mode_ru     => 'усилия/показа',
    mode1_ru    => 'послед./подробно', mode2_ru => 'все/обзор с',
    hours_en    => 'hour(s)',
    hours_de    => 'Stunde(n)',
    hours_es    => 'hora(s)',
    hours_ru    => 'час(ов)',
    decode_en   => 'decode',
    decode_de   => 'Übersetzen',
    decode_es   => 'traduce',
    decode_ru   => 'переводить',
    disclaimer_en => 'Warning: The information on this page may be out-dated, inaccurate, or both. It is not suited for use in aviation.',
    disclaimer_de => 'Warnung: Die Informationen auf dieser Seite könnten veraltet, falsch oder beides sein. Sie sind nicht für die Verwendung in der Luftfahrt geeignet.',
    disclaimer_es => 'Aviso: Las informaciones de esta página web puedan ser antiguas o falsas o los dos. No son apropiadas para el uso en la aviación.',
    disclaimer_ru => 'Предупреждение: Информация относительно этой страницы может быть устарелой, неточной, или оба. Этому не удовлетворяют для использования в авиации.',
);

#
# save message parameter
#
my $msg_metaf_orig = $msg_metaf;
my $msg_synop_orig = $msg_synop;

# 2 blanks are used as delimiter between msg_metaf and msg_synop for -O
$msg_metaf_orig =~ s/  +/ /gs;
$msg_metaf_orig =~ s/ +$//s;

=pod

If C<type_metaf> is B<icao> and one or more airport ICAO codes are given,
messages for the specified airports are fetched from the local file
F<$CGI_DATA_DIR/metaf.txt> (if it exists) or the specified Internet server.

If C<type_synop> is B<wmo> and one or more WMO ids are given,
messages for the specified stations are fetched from the local file
F<$CGI_DATA_DIR/synop.txt> (if it exists) or the specified Internet server.

=cut

#
# process METAR/TAF messages if they need to be fetched
#
if (   $type_metaf eq 'icao'
    && " $msg_metaf" =~ /^(?: (?:$re_report_type_metaf-)?$re_ICAO)+$/)
{
    my ($icao_uniq, $icao_uniq_metar, $icao_uniq_staf, $icao_uniq_taf, @metaf);

    $icao_uniq = mk_uniq $msg_metaf =~ /(?:$re_report_type_metaf-)?([^ ]+)/og;
    $icao_uniq_metar = '';
    $icao_uniq_staf  = '';
    $icao_uniq_taf   = '';
    for ($msg_metaf =~ /((?:$re_report_type_metaf-)?(?:[^ ]+))/og) {
        /($re_report_type_metaf-)?(.*)/;
        if (defined $1) {
            if ($1 eq 'SA-' || $1 eq 'SP-') {
                if ($icao_uniq_metar eq '') {
                    $icao_uniq_metar = "$1$2";
                } elsif (index(" $icao_uniq_metar ", " $1$2 ") == -1) {
                    $icao_uniq_metar .= " $1$2";
                }
            } elsif ($1 eq 'FC-') {
                if ($icao_uniq_staf eq '') {
                    $icao_uniq_staf = "$1$2";
                } elsif (index(" $icao_uniq_staf ", " $1$2 ") == -1) {
                    $icao_uniq_staf .= " $1$2";
                }
            } else { # FT
                if ($icao_uniq_taf eq '') {
                    $icao_uniq_taf = "$1$2";
                } elsif (index(" $icao_uniq_taf ", " $1$2 ") == -1) {
                    $icao_uniq_taf .= " $1$2";
                }
            }
        } else {
            if ($icao_uniq_metar eq '') {
                $icao_uniq_metar = "SA-$2 SP-$2";
            } else {
                if (index(" $icao_uniq_metar ", " SA-$2 ") == -1) {
                    $icao_uniq_metar .= " SA-$2";
                }
                if (index(" $icao_uniq_metar ", " SP-$2 ") == -1) {
                    $icao_uniq_metar .= " SP-$2";
                }
            }
            if ($icao_uniq_staf eq '') {
                $icao_uniq_staf = "FC-$2";
            } elsif (index(" $icao_uniq_staf ", " FC-$2 ") == -1) {
                $icao_uniq_staf .= " FC-$2";
            }
            if ($icao_uniq_taf eq '') {
                $icao_uniq_taf = "FT-$2";
            } elsif (index(" $icao_uniq_taf ", " FT-$2 ") == -1) {
                $icao_uniq_taf .= " FT-$2";
            }
        }
    }

    if ($src_metaf eq 'noaa') {
        if ($mode eq 'summary') {
            @metaf = (search_files_metaf($icao_uniq_metar, $mode,
                                         'metar_cycle_iws', $hours),
                      search_files_metaf($icao_uniq_staf, $mode,
                                         'staf_cycle_iws', $hours),
                      search_files_metaf($icao_uniq_taf, $mode,
                                         'taf_cycle_iws', $hours));
        } else {
            # faster than search_files_metaf(), but may return outdated messages
            @metaf = (query_metaf_iws($icao_uniq_metar, 'metar', $mode),
                      query_metaf_iws($icao_uniq_staf,  'staf',  $mode),
                      query_metaf_iws($icao_uniq_taf,   'taf',   $mode));
        }
    } elsif ($src_metaf eq 'adds') {
        @metaf = query_metaf_adds $icao_uniq, $mode, $hours,
                                  "$icao_uniq_staf$icao_uniq_taf" ne '';
    } elsif ($src_metaf eq 'addsds') {
        @metaf = (query_metaf_addsds(mk_uniq($icao_uniq_metar
                                      =~ /(?:$re_report_type_metaf-)([^ ]+)/og),
                                     'metar', $mode, $hours),
                  query_metaf_addsds(mk_uniq("$icao_uniq_staf $icao_uniq_taf"
                                      =~ /(?:$re_report_type_metaf-)([^ ]+)/og),
                                     'taf', $mode, $hours));
    } elsif ($src_metaf eq 'ogimet') {
        my $type = 'ALL';
        $type = $1
            if $msg_metaf =~ /^($re_report_type_metaf)-[^ ]+(?: \1-[^ ]+)*$/;
        @metaf = query_metaf_ogimet $icao_uniq, $mode, $hours, $type;
    } elsif ($src_metaf eq 'cod' || $src_metaf eq 'fsu') {
        @metaf = search_files_metaf $icao_uniq_metar, $mode, $src_metaf, $hours;
    } elsif ($src_metaf eq 'nws') {
        @metaf = (search_files_metaf($icao_uniq_metar,$mode,'nws_metar',$hours),
                  search_files_metaf(
                    "$icao_uniq_staf $icao_uniq_taf", $mode,'nws_taf', $hours));
    } else {
        @metaf = search_files_metaf
                              "$icao_uniq_metar $icao_uniq_staf $icao_uniq_taf",
                                    $mode, $src_metaf, $hours;
    }

    $msg_metaf = '';
    for (@metaf) {
        if (/ERROR/) {
            $msg_metaf = $_;
            last;
        }
    }

    if (!$msg_metaf) {
        for my $id (split / /, $icao_uniq) {
            my ($id_list, $ii, $type_pattern);

            $ii = 0;
            $ii++    if index(" $icao_uniq_metar ", " SA-$id ") > -1;
            $ii += 2 if index(" $icao_uniq_metar ", " SP-$id ") > -1;
            if ($ii > 0) {
                $type_pattern = qw(SA SP S[AP])[$ii - 1];
                $id_list = join("\n", grep { /^$type_pattern $id / } @metaf);
                if ($id_list) {
                    $id_list =~ s/^SA //mg;
                    $id_list =~ s/^SP/SPECI/mg;
                    $msg_metaf .= "\nMETAR\n" . $id_list;
                }
            }
            $ii = 0;
            $ii++    if index(" $icao_uniq_staf ", " FC-$id ") > -1;
            $ii += 2 if index(" $icao_uniq_taf ",  " FT-$id ") > -1;
            if ($ii > 0) {
                $type_pattern = qw(FC FT F[CT])[$ii - 1];
                $id_list = join("\n", grep { /^$type_pattern (?:COR |AMD )*$id / } @metaf);
                if ($id_list) {
                    $id_list =~ s/^F[CT] //mg;
                    $msg_metaf .= "\nTAF\n" . $id_list;
                }
            }
        }
    }
} elsif ($msg_metaf ne '') {
    $msg_metaf = ($type_metaf eq 'taf' ? 'TAF' : 'METAR') . "\n$msg_metaf";
}

# process SYNOP messages if they need to be fetched
if (   $type_synop eq 'wmo'
    && " $msg_synop" =~ /^(?: (?:$re_report_type_synop-)?[A-Z\d]{3,})+$/)
{
    my ($wmoid_uniq, @synop);

    $wmoid_uniq = mk_uniq $msg_synop =~ /(?:$re_report_type_synop-)?([^ ]+)/og;
    if ($src_synop eq 'ogimet') {
        my $type = 'ALL';
        $type = $1
            if $msg_synop =~ /^($re_report_type_synop)-[^ ]+(?: \1-[^ ]+)*$/;
        @synop = query_synop_ogimet $msg_synop, $mode, $wmoid_uniq, $hours,
                                    $type;
    } else {
        @synop = search_files_synop $msg_synop, $mode, $src_synop, $hours;
    }

    $msg_synop = '';
    for (@synop) {
        if (/ERROR/) {
            $msg_synop = $_;
            last;
        }
    }

    if (!$msg_synop) {
        for my $id (split / /, $wmoid_uniq) {
            $msg_synop .=
                join("\n",
                 grep { /^$re_report_type_synop $re_header_synop $id / } @synop)
            . "\n";
        }
        if ($msg_synop) {
            $msg_synop =~ s/^$re_report_type_synop //mog;
            $msg_synop = "\nSYNOP\n" . $msg_synop;
        }
    }
} elsif ($msg_synop ne '') {
    $msg_synop = 'SYNOP' . "\n$msg_synop";
}

my $msg = $msg_metaf . "\n" . $msg_synop;
$msg =~ s/ +/ /g;
$msg =~ s/ $//g;
$msg =~ s/[^\n -~]/?/gos; # avoid invalid XML and HTML
$msg =~ s/[\n]+/\n/g;
$msg =~ s/^\n//;
$msg =~ s/\n$//;

if ($format eq 'xml') {
    print header({type => 'application/xml'})
        unless $is_cmdline;
    my $xsl = $mode eq 'summary' ? 'sum-ui' : 'fullhtml';
    open(METAF2XML, '|-', "$METAF2XML_BIN/metaf2xml.pl", '-Xx-', "-S../metaf-$xsl.xsl", "-O$type_metaf $type_synop $lang $format $src_metaf $src_synop $mode $hours $msg_metaf_orig  $msg_synop_orig");
    print METAF2XML $msg;
    close METAF2XML;
    exit 0;
}

print header()
    unless $is_cmdline;
print start_html({ title => $trans{'title_'.$lang},
                   lang  => $lang,
                   dtd   => [ '-//W3C//DTD HTML 4.01 Transitional//EN',
                              'http://www.w3.org/TR/html4/loose.dtd' ],
                   meta  => { robots   => 'noindex,nofollow',
                              keywords =>
               'metaf2xml,METAR,TAF,SYNOP,XML,aviation,weather,report,forecast',
                          },
                   head  => meta({ http_equiv => 'Content-Type',
                                   content    => 'text/html; charset=UTF-8'})
               }
              ) . "\n" .
    h2($trans{'header_'.$lang}) . "\n"
        if $need_html;

print
    $trans{'infosrc_'.$lang} . ":\n" .
    kbd('http://weather.noaa.gov/weather/coded.html') . p('') . "\n" .
    $trans{'infofetch_'.$lang} . p('') . "\n" .
    start_form({ action => 'metaf.pl' }) .
    table({ border => 0, cellspacing => 0 }, "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'language_'.$lang} . ': ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'de',
                      ($lang eq 'de' ? 'checked' : '') => undef }
                   ) .
              'Deutsch ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'en',
                      ($lang eq 'en' ? 'checked' : '') => undef }
                   ) .
              'English ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'es',
                      ($lang eq 'es' ? 'checked' : '') => undef }
                   ) .
              'Español ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'ru',
                      ($lang eq 'ru' ? 'checked' : '') => undef }
                   ) .
              'Русский ') . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'format_'.$lang} . ': ') . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'html',
                      ($format eq 'html' ? 'checked' : '') => undef }
                   ) .
              $trans{'format1_'.$lang} . ' ') . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'text',
                      ($format eq 'text' ? 'checked' : '') => undef }
                   ) .
              $trans{'format2_'.$lang} . ' ') . "\n" .
           td({ colspan => 2 },
              input({ type => 'radio', name => 'format', value => 'xml',
                      ($format eq 'xml' ? 'checked' : '') => undef }
                   ) .
              $trans{'format3_'.$lang} . ' ') . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans{'mode_'.$lang} . ': ') . "\n" .
           td({ nowrap => undef },
              input({ type => 'radio', name => 'mode', value => 'latest',
                      ($mode eq 'latest' ? 'checked' : '') => undef }
                   ) .
              $trans{'mode1_'.$lang} . ' ') . "\n" .
           td({ colspan => 3, nowrap => undef },
              input({ type => 'radio', name => 'mode', value => 'summary',
                      ($mode eq 'summary' ? 'checked' : '') => undef }
                   ) .
              $trans{'mode2_'.$lang} . "\n" .
              textfield({ name      => 'hours',
                          size      => 3,
                          maxlength => 3,
                          style   => 'font-family: monospace; font-size: 100%;',
                          default   => $hours,
                          override  => 1}) . ' ' .
              $trans{'hours_'.$lang} . ' '
             ) . "\n"
        ) . "\n"
    ) . "\n" .
    table({ border => 0, cellspacing => 0 }, "\n" .
       Tr(td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top' }, "\n" .
             $trans{'example_'.$lang} . ': ') . "\n" .
          td({ style => 'font-family: monospace;' }, "\n" .
             'METAR YUDO 090600Z 00000KT CAVOK 22/15 Q1021 NOSIG' . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
          td('') . "\n" .
          td({ style => 'font-family: monospace;' }, "\n" .
             'TAF YUDO 090600Z 0907/0916 VRB03KT 8000 SKC PROB40 TEMPO 0907/0908 0200 +TSRA FM090800 CAVOK' . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
          td('') . "\n" .
          td({ style => 'font-family: monospace;' }, "\n" .
             'TAF YUDO 090600Z 090716 VRB03KT 8000 SKC PROB40 TEMPO 0708 0200 +TSRA FM0800 CAVOK' . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_metaf', value => 'metar',
                   ($type_metaf eq 'metar' ? 'checked' : '') => undef }
                ) .
           $trans{'type1_'.$lang} . " \n"
       ). "\n" .
        td({ rowspan => 3, valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_metaf',
                       size      => 80,
                       maxlength => 500,
                       style     => 'font-family: monospace; font-size: 100%;',
                       default   => $msg_metaf_orig,
                       override  => 1}) . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef },
           input({ type => 'radio', name => 'type_metaf', value => 'taf',
                   ($type_metaf eq 'taf' ? 'checked' : '') => undef }
                ) .
           $trans{'type2_'.$lang} . " \n"
          ). "\n"
       ). "\n" .
       Tr("\n" .
        td({ nowrap => undef },
           input({ type => 'radio', name => 'type_metaf', value => 'icao',
                   ($type_metaf eq 'icao' ? 'checked' : '') => undef }
                ) .
           $trans{'type4_'.$lang} . ": \n"
          ). "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans{'source_'.$lang} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_metaf', value => 'noaa',
                     ($src_metaf eq 'noaa' ? 'checked' : '') => undef,
                     ($src_metaf eq 'file' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source1_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'adds',
                     ($src_metaf eq 'adds' ? 'checked' : '') => undef,
                     ($src_metaf eq 'file' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source2_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'addsds',
                     ($src_metaf eq 'addsds' ? 'checked' : '') => undef,
                     ($src_metaf eq 'file' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source3_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'ogimet',
                     ($src_metaf eq 'ogimet' ? 'checked' : '') => undef,
                     ($src_metaf eq 'file' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source6_'.$lang} . ' '
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top' }, "\n" .
             $trans{'example_'.$lang} . ': ') . "\n" .
          td({ style => 'font-family: monospace;' }, "\n" .
             'AAXX 09004 08495 11459 30714 10147 20136 30151 40159 58005 60001 70511 83500 92350' . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
          td('') . "\n" .
          td({ style => 'font-family: monospace;' }, "\n" .
             'BBXX UIDS 18061 99613 70284 41698 61806 10000 40107 57020 70121 86//8 22271 00064' . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_synop', value => 'synop',
                   ($type_synop eq 'synop' ? 'checked' : '') => undef }
                ) .
           $trans{'type3_'.$lang} . " \n"
       ). "\n" .
        td({ rowspan => 2, valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_synop',
                       size      => 80,
                       maxlength => 500,
                       style     => 'font-family: monospace; font-size: 100%;',
                       default   => $msg_synop_orig,
                       override  => 1}) . "\n") . "\n"
       ) . "\n" .
       Tr("\n" .
        td(input({ type => 'radio', name => 'type_synop', value => 'wmo',
                   ($type_synop eq 'wmo' ? 'checked' : '') => undef }
                ) .
           $trans{'type5_'.$lang} . ": \n"
          ). "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans{'source_'.$lang} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_synop', value => 'nws_land',
                     ($src_synop eq 'nws_land' ? 'checked' : '') => undef,
                     ($src_synop eq 'file' ? 'readonly' : '') => undef,
                     ($src_synop eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source4_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_synop', value => 'nws_ship',
                     ($src_synop eq 'nws_ship' ? 'checked' : '') => undef,
                     ($src_synop eq 'file' ? 'readonly' : '') => undef,
                     ($src_synop eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source7_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_synop', value => 'cod',
                     ($src_synop eq 'cod' ? 'checked' : '') => undef,
                     ($src_synop eq 'file' ? 'readonly' : '') => undef,
                     ($src_synop eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source5_'.$lang} . "\n" .
             input({ type => 'radio', name => 'src_synop', value => 'ogimet',
                     ($src_synop eq 'ogimet' ? 'checked' : '') => undef,
                     ($src_synop eq 'file' ? 'readonly' : '') => undef,
                     ($src_synop eq 'file' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans{'source6_'.$lang} . ' '
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ colspan => 2 },
           submit({ style => 'font-size: 100%;',
                    value => $trans{'decode_'.$lang}}))
          . "\n")
    ) . "\n" .
    end_form . p('') . "\n"
        unless $is_cmdline;

if ($msg_metaf or $msg_synop) {
    my $response;
    my $sum = $mode eq 'summary' ? 'sum-' : '';

    if ($need_html) {
        print b($trans{'disclaimer_'.$lang}), p(''), "\n";
    } else {
        print $trans{'disclaimer_'.$lang}, "\n\n";
    }

    if (!open(METAF2XML,   "| $METAF2XML_BIN/metaf2xml.pl -Xx-"
                         . "| xsltproc --nonet --nowrite --nomkdir --stringparam lang $lang --stringparam wrap_table $need_html $METAF2XML_XSL/metaf-$sum$format.xsl -"))
    {
        print p(''), b("ERROR starting $METAF2XML_BIN/metaf2xml.pl: $!"), p('');
    } else {
        print METAF2XML $msg;
        if (!close METAF2XML) {
            print p(''), b("ERROR starting xsltproc"), p('');
        }
    }
}

if ($need_html) {
    print comment('metaf.pl: 1.40'), "\n", p(''),
          "copyright (c) 2006-2010 metaf2xml @\n",
          a({ href => 'http://metaf2xml.sourceforge.net/' }, '
            SourceForge.net
          '),
          end_html();
} else {
    print 'metaf.pl: 1.40',
          " copyright (c) 2006-2010 metaf2xml @ SourceForge.net\n";
}

=head1 EXAMPLES

Parse a METAR message and print the result as HTML:

  metaf.pl 'msg_metaf=YUDO 090600Z 00000KT CAVOK 22/15 Q1021'

Fetch up-to-date METAR and TAF messages (from NOAA/IWS) and SYNOP messages (from
NOAA/NWS land) for Rio (C<SBGL>, 83746), New York (C<KJFK>, 74486), and Tokyo
(C<RJTT>, 47662) and print the result as XML:

  metaf.pl 'format=xml&type_metaf=icao&type_synop=wmo&msg_metaf=SBGL KJFK RJTT&msg_synop=83746 74486 47662'

=head1 SEE ALSO

=begin html

<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metafsrc2raw.pl.html>metafsrc2raw(1)</a>, <!--

=end html

B<metafsrc2raw>(1),
B<metaf2xml>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2006-2010 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut
